#!/usr/local/bin/ruby

# rw-cgi.rb -- RWiki CGI interface.
#
# $Id: rw-cgi.rb,v 1.44 2002/10/15 15:11:47 seki Exp $
# Copyright (c) 2000-2001 Masatoshi SEKI and NAKAMURA, Hiroshi
#
# rw-cgi.rb is copyrighted free software by Masatoshi SEKI and NAKAMURA, Hiroshi.
# You can redistribute it and/or modify it under the same term as Ruby.


$KCODE = 'EUC'	# SETUP
$RWIKI_DIR = File::dirname( ENV['SCRIPT_FILENAME'] )	# SETUP
$RWIKI_URI = 'druby://localhost:8470'			# SETUP


$:.unshift( $RWIKI_DIR )
$:.unshift( File::join( $RWIKI_DIR, 'lib' ))


###
# SYNOPSIS
#   CGIApp.new( appName )
#
# ARGS
#   appName     Name String of the CGI application.
#
# DESCRIPTION
#   Running this application object as a CGI with 'cmd=[value]' causes
#   executing the method 'exec_proc_[method]_[value]'.
#     ie. POST to 'cmd=entry' -> exec_proc_post_entry
#     ie. HEAD to 'cmd=view' -> exec_proc_head_view
#   Without 'cmd=[value]' in query, 'exec_proc_default' are called.
#
require 'devel/logger'
require 'cgi'
class CGIApp < Devel::Application
  include Devel::Logger::Severity

  public

  def initialize( appName )
    super( appName )
    @cgi = CGI.new()
    @query = @cgi.params
    @cookies = @cgi.cookies
    @os = @cgi
  end

  private

  def run
    prologue()

    log( SEV_INFO, "CGIApp#run: Accessed user '" <<
      ( @cgi.remote_user || 'anonymous' ) << '@' <<
      ( @cgi.remote_host || @cgi.remote_addr || 'unknown' ) << "'." )
    log( SEV_DEBUG, "The query invoked with: #{ @cgi.inspect }." )

    begin
      response = nil
      if ( @query.has_key?( 'cmd' ))
	msg = "exec_proc_#{ @cgi.request_method.downcase }_#{ @query[ 'cmd' ][0] }"
	if self.respond_to?( msg, true )
	  response = send( msg.intern )
	end
      end

      unless response
	msg = "exec_proc_#{ @cgi.request_method.downcase }"
	if self.respond_to?( msg, true )
	  response = send( msg.intern )
	end
      end

      unless response
	response = exec_proc_default
      end

      str = response.dump
      log( SEV_DEBUG, "Response:\n#{ str }" )
      print str
    rescue Errno::ECONNREFUSED
      errMessage = 'RWiki server seems to be down...'
      log( SEV_DEBUG, "Response:\n#{ response }" )
      print Response.new( Response::Header.new( 400 ), Response::Body.new( errMessage )).dump
      raise
    rescue
      errMessage = 'CGI exec failed.'
      log( SEV_DEBUG, "Response:\n#{ response }" )
      print Response.new( Response::Header.new( 400 ), Response::Body.new( errMessage )).dump
      raise
    end

    epilogue()

    return 0
  end

  def prologue; end
  def epilogue; end

  class Response

    public

    class Header
      attr_accessor :status, :bodyType, :bodyCharset, :bodySize, :bodyDate

      CRLF = "\r\n"

      StatusMap = {
	200 => 'OK',
	302 => 'Object moved',
	400 => 'Bad Request',
      }

      CharsetMap = {
	'NONE' => 'us-ascii',
	'EUC' => 'euc-jp',
	'SJIS' => 'shift_jis',
      }

      def initialize( status = 200 )
	@status = status
	@bodyType = nil
	@bodyCharset = nil
	@bodySize = nil
	@bodyDate = nil
	@extra = []
      end

      def add( key, value )
	@extra.push( [ key, value ] )
      end

      def dump
	str = ''
	if defined?( Apache )
	  if !StatusMap.include?( @status )
	    @status = 400
	  end
	  str << dumpItem( "HTTP/1.0 #{ @status } #{ StatusMap[ @status ] }" )
	  str << dumpItem( "Date: #{ httpDate( Time.now ) }" )
	end

	if @bodySize
	  str << dumpItem( "Content-Length: #{ @bodySize }" )
	else
	  str << dumpItem( "Connection: close" )
	end
	str << dumpItem( "Last-Modified: #{ httpDate( @bodyDate ) }" ) if @bodyDate
	str << dumpItem( "Content-Type: #{ @bodyType || 'text/html' }; charset=#{ CharsetMap[ @bodyCharset || $KCODE ]}" )
	@extra.each do | key, value |
	  str << dumpItem( "#{ key }: #{ value }" )
	end
	str << CRLF
	str
      end

      private

      def dumpItem( str )
	str + CRLF
      end

      def httpDate( aTime )
	aTime.gmtime.strftime( "%a, %d %b %Y %H:%M:%S GMT" )
      end
    end

    class Body
      attr_accessor :type, :charset, :date

      def initialize( body = nil, date = nil, type = nil, charset = nil )
	@body = body
	@type = type
	@charset = charset
	@date = date
      end

      def size
	if @body
	  @body.size
	else
	  nil
	end
      end

      def dump
	@body
      end
    end

    def initialize( header = nil, body = nil )
      self.header = header
      self.body = body
    end

    def dump
      unless header
	raise RuntimeError.new( "Response header not set." )
      end
      str = header.dump
      str << body.dump if body and body.dump
      str
    end

    def header
      @header
    end

    def header=( header )
      @header = header
      sync
    end

    def body
      @body
    end

    def body=( body )
      @body = body
      sync
    end

    private

    def sync
      if @header and @body
	@header.bodyType = body.type
	@header.bodyCharset = body.charset
	@header.bodySize = body.size
	@header.bodyDate = body.date
      end
    end
  end
end


###
# SYNOPSIS
#   RWikiCGIApp.new( rwiki )
#
# ARGS
#   rwiki     DRb::DRbObject of RWiki server.
#
# DESCRIPTION
#   RWiki CGI interface.
#
require 'drb/drb'
require 'rw-lib'
class RWikiCGIApp < CGIApp
  include Devel::Logger::Severity

  VERSION = ['rw-cgi', '1.2.4 ext']

  public

  LogDir = File::join( $RWIKI_DIR, 'log' )

  def initialize( rwiki )
    super( AppName )
    @rwiki = rwiki
    setLog( File::join( LogDir, AppName + '.log' ), 'weekly' )
  end

  private

  AppName = 'RWikiCGIApp'

  def prologue
    @log.sevThreshold = SEV_INFO
  end

  def exec_proc_head_view
    begin
      req = parseRequest()
      raise RWInvalidRequest unless req.name
      page = @rwiki.page( req.name )
      Response.new( Response::Header.new(), Response::Body.new())
    rescue RWiki::InvalidRequest
      requestError
    rescue
      unknownError( $!, $@ )
    end
  end

  def exec_proc_get_view
    begin
      req = parseRequest()
      raise RWInvalidRequest unless req.name
      page = @rwiki.page( req.name )
      header = Response::Header.new()
      if page.empty?
	res = page.edit_html( getEnv ) { |key| @query[key] }
	header.add( 'Cache-Control', 'private' )
      elsif @query.has_key?( 'em' )
	res = page.emphatic_html( getEnv ) { |key| @query[key] }
      else
	res = page.view_html( getEnv ) { |key| @query[key] }
      end
      Response.new( header, Response::Body.new( res ))
    rescue RWiki::InvalidRequest
      requestError
    rescue RWiki::RevisionError
      revisionError( req )
    rescue
      unknownError( $!, $@ )
    end
  end

  def exec_proc_get_edit
    begin
      req = parseRequest()
      page = @rwiki.page( req.name )
      header = Response::Header.new()
      header.add( 'Cache-Control', 'private' )
      res = page.edit_html( getEnv ) { |key| @query[key] }
      Response.new( header, Response::Body.new( res ))
    rescue RWiki::InvalidRequest
      requestError
    rescue
      unknownError( $!, $@ )
    end
  end

  def exec_proc_head_src
    begin
      req = parseRequest()
      raise RWInvalidRequest unless req.name
      modified = @rwiki.page( req.name ).modified
      if ENV.key?('HTTP_IF_MODIFIED_SINCE') && modified <= Time.parse(ENV['HTTP_IF_MODIFIED_SINCE'])
        return Response.new( Response::Header.new( 304 ) )
      end
      Response.new( Response::Header.new(), Response::Body.new( nil, modified ))
    rescue RWiki::InvalidRequest
      requestError
    rescue
      unknownError( $!, $@ )
    end
  end

  def exec_proc_get_src
    begin
      req = parseRequest()
      raise RWInvalidRequest unless req.name
      page = @rwiki.page( req.name )
      modified = page.modified
      if ENV.key?('HTTP_IF_MODIFIED_SINCE') && modified <= Time.parse(ENV['HTTP_IF_MODIFIED_SINCE'])
        return Response.new( Response::Header.new( 304 ) )
      end
      res = @rwiki.src_view( req.name, getEnv ) { |key|
          case key
          when 'message'
            page.src
          else
            ''
          end
        }
      Response.new( Response::Header.new(), Response::Body.new( res, modified ))
    rescue RWiki::InvalidRequest
      requestError
    rescue
      unknownError( $!, $@ )
    end
  end

  def exec_proc_post_submit
    begin
      req = parseRequest()
      raise RWInvalidRequest unless req.name
      raise RWInvalidRequest unless req.src
      page = @rwiki.page( req.name )
      page.set_src( req.src, req.rev )
      header = Response::Header.new()
      res = page.submit_html( getEnv ) { |key| @query[key] }
      Response.new( header, Response::Body.new( res ))
    rescue RWiki::InvalidRequest
      requestError
    rescue RWiki::RevisionError
      revisionError( req )
    rescue
      unknownError( $!, $@ )
    end
  end

  alias exec_proc_default exec_proc_get_view

  def revisionError( req )
    errMessage =<<__EOM__
Source revision mismatch.

Page '#{ req.name }' has changed since editing started.
Back to the edit page and press 'RELOAD' to refresh the page,
then retry to merge/add your changes to its latest source.

Submitted source:
#{ req.src }
__EOM__
    errorResponse( req.name, errMessage )
  end

  def requestError
    redirectTo = RWiki::Request.default_url( ENV )
    log( SEV_INFO, "Redirect to '#{ redirectTo }'." )
    header = Response::Header.new( 302 )
    header.add( 'Location', redirectTo )
    Response.new( header )
  end

  def unknownError( err, info )
    errMessage = "#{ err }\n\n" << info.join( "\n" )
    errorResponse( 'error', errMessage )
  end

  def errorResponse( name, msg )
    name = 'error' unless name
    res = @rwiki.error_view( name, getEnv ) { |key|
	case key
	when 'message'
	  msg
	else
	  ''
	end
      }
    Response.new( Response::Header.new( 400 ), Response::Body.new( res ))
  end

  def parseRequest
    log( SEV_DEBUG, "REQUEST_URI: #{ ENV[ 'REQUEST_URI' ] }" )
    log( SEV_DEBUG, "SCRIPT_NAME: #{ ENV[ 'SCRIPT_NAME' ] }" )
    req = RWiki::Request.parse( @query )
    log( SEV_INFO,  "Request: #{ req.inspect }" )
    req
  end

  def getEnv
    env = Hash.new
    env[ 'base' ] = ENV[ 'SCRIPT_NAME' ]
    env[ 'base_url' ] = RWiki::Request.base_url(ENV)
    env[ 'server' ] = ENV[ 'SERVER_NAME' ] + ((ENV['SERVER_PORT'] == '80') ? '' : ':' + ENV['SERVER_PORT'])
    env[ 'rw-agent-info' ] = [ VERSION ]
# Recover these comment-out-ed lines and this proc will generates URL of
# RWiki name's references.  But bare in mind Proc cannot be dumped and Hash
# containing Proc cannot be dumped so a dRuby connection is needed per a
# reference generation.
#    env[ 'ref_name' ] = Proc.new { |cmd, name|
#	program = env[ 'base' ]
#	req = RWiki::Request.new(cmd, name)
#	program + "?" + req.query
#      }
    env
  end
end


DRb.start_service()
$rwiki = DRbObject.new( nil, $RWIKI_URI )
$app = RWikiCGIApp.new( $rwiki ).start()
